
/*
 * Web Polygraph                        http://www.web-polygraph.org/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. See the COPYRIGHT file for details.
 */

#include "base/polygraph.h"

#include <ctype.h>
#include "xstd/h/iostream.h"
#include "xstd/h/iomanip.h"

#include "base/ObjId.h"
#include "base/RndPermut.h"
#include "base/polyLogCats.h"
#include "base/polyLogTags.h"
#include "runtime/AddrMap.h"
#include "runtime/HostMap.h"
#include "runtime/Viserv.h"
#include "runtime/PubWorld.h"
#include "runtime/StatPhase.h"
#include "runtime/StatPhaseMgr.h"
#include "runtime/LogComment.h"
#include "runtime/ErrorMgr.h"
#include "runtime/EphPortMgr.h"
#include "runtime/ExpPortMgr.h"
#include "runtime/PopModel.h"
#include "runtime/polyBcastChannels.h"
#include "runtime/polyErrors.h"
#include "runtime/StatPhaseMgr.h"
#include "runtime/globals.h"
#include "runtime/Viserv.h"
#include "csm/ContentSel.h"
#include "csm/ContentCfg.h"
#include "csm/ContentMgr.h"
#include "pgl/RobotSym.h"
#include "pgl/PglSemx.h"
#include "dns/DnsMgr.h"
#include "client/CltConnMgr.h"
#include "client/CltOpts.h"
#include "client/CltXact.h"
#include "client/IcpCltXact.h"
#include "client/SessionMgr.h"
#include "client/ServerRep.h"
#include "client/PrivCache.h"
#include "client/CltCfg.h"
#include "client/Client.h"


class IcpAgentCfg;

struct SharedCfgItem { const RobotSym *sym; CltCfg *cfg; };
static Array<SharedCfgItem> SharedCfgs;

XactFarm<CltXact> *Client::TheXacts = 0;
ObjFarm<IcpCltXact> Client::TheIcpXacts;

CltSharedCfgs *Client::TheSharedCfgs = new CltSharedCfgs;
Array<PortMgr*> Client::ThePortMgrs;


Client::Client(): thePrivCache(0), theCfg(0), 
	theConnMgr(0), theDnsMgr(0), theSessionMgr(0), theIcpClient(0), 
	theCcXactLvl(0), theExtraLaunchLvl(0), isIdle(false) {

	theChannels.append(TheInfoChannel);
	theChannels.append(TheLogCfgChannel);
	theChannels.append(TheLogStateChannel);
	startListen();
}

Client::~Client() {
	delete theConnMgr;
	delete theDnsMgr;
	delete theSessionMgr;
	delete thePrivCache;
}

void Client::configure(const RobotSym *cfg, const NetAddr &aHost) {
	Assert(TheXacts);
	Assert(aHost.port() < 0); // remove later

	SockOpt opt;
	Agent::configure(cfg, aHost, opt);

	theCfg = TheSharedCfgs->getConfig(cfg);

	thePrivWorlds.stretch(theCfg->originCount());
	for (int i = 0; i < theCfg->originCount(); ++i)
		thePrivWorlds.append(PrivWorld(theId));

	int pcCap = 0;
	if (cfg->privCache(pcCap))
		thePrivCache = new PrivCache(pcCap);

	theCfg->selectProxy(theProxyAddr); // sticky selection

	theConnMgr = new CltConnMgr;
	theConnMgr->configure(opt, cfg, theProxyAddr ? 1 : theCfg->originCount());
	theConnMgr->portMgr(getPortMgr(true));
	theConnMgr->idleTimeout(cfg->idlePconnTimeout());

	theDnsMgr = new DnsMgr(this);
	theDnsMgr->configure(cfg->dnsResolver());

	if (theCfg->theBusyPeriod) {
		theSessionMgr = new SessionMgr(this);
		theSessionMgr->configure(theCfg);
	}

	if (theCfg->thePeerHttp && !theCfg->thePeerIcp) {
		cerr << cfg->loc() << "the HTTP peer is at " << theCfg->thePeerHttp << ", but where is the ICP peer?" << endl;
		exit(-3);
	}

	TheXacts->limit(1024); // magic, no good way to estimate
}

void Client::icpClient(IcpClient *anIcpClient) {
	Assert(!theIcpClient);
	theIcpClient = anIcpClient;
}

void Client::start() {
	Assert(theConnMgr);
	Assert(theDnsMgr);

	Agent::start();

	theDnsMgr->start();

	theCfg->startWarmup(); // warmup plan created if needed

	if (theSessionMgr)
		theSessionMgr->start();
	else
		becomeBusy();
}

void Client::stop() {
	if (theSessionMgr)
		theSessionMgr->stop();
	else
	if (!isIdle)
		becomeIdle();

	theDnsMgr->stop();
	Agent::stop();
}

void Client::becomeBusy() {
	isIdle = false;
	theCfg->selectUsername(theUsername);
	Broadcast(TheSessionBegChannel, this);
	scheduleLaunch(TheClock);
}

void Client::continueSession() {
	Broadcast(TheSessionCntChannel, this);
}

void Client::becomeIdle() {
	isIdle = true;

	theLaunchDebts.reset();
	theExtraLaunchLvl = 0;

	theConnMgr->closeAllIdle();

	Broadcast(TheSessionEndChannel, this);
}

void Client::describe(ostream &os) const {
	Agent::describe(os);
	if (theProxyAddr)
		os << " via " << theProxyAddr;
}

void Client::noteInfoEvent(BcastChannel *ch, InfoEvent ev) {
	Assert(ch == TheInfoChannel);
	if (ev == ieWssFreeze) {
		for (int i = 0; i < thePrivWorlds.count(); ++i)
			thePrivWorlds[i].freezeWss();

		static bool frozen = false;
		if (!frozen) {
			FreezeSrvWss();
			frozen = true;
			const BigSize meanFillSz = TheStatPhaseMgr->fillCnt() ?
				TheStatPhaseMgr->fillSz()/TheStatPhaseMgr->fillCnt() :
				BigSize(0);
			const int pubWss = PubWorld::PositiveWss();
			const int privWss = PrivWorld::TheWss;
			Comment(5) << "fyi: min `direct' objects in working set:"
				<< " global public: "
					<< pubWss << " (" << (meanFillSz*pubWss) << ')'
				<< " local private: "
					<< privWss << " (" << (meanFillSz*privWss) << ')'
				<< endc;
		}
	}
}

void Client::noteLogEvent(BcastChannel *ch, OLog &log) {
	if (ch == TheLogCfgChannel) {
		log << bege(lgCltCfg, lgcCltSide);
		// XXX: implement (CltCfgRec)
		log << ende;
	} else
	if (ch == TheLogStateChannel) {
		log << bege(lgCltState, lgcCltSide)
			<< theSeqvId 
			<< thePrivWorlds
			<< ende;
	}
}

void Client::noteXactDone(CltXact *x) {
	Assert(x);

	const ObjId oid(x->oid());
	const bool retry = x->needRetry();

	if (theSessionMgr)
		theSessionMgr->noteXactDone(x);

	theConnMgr->put(x->conn());
	TheXacts->put(x);
	theCcXactLvl--;
	Assert(theCcXactLvl >= 0);

	// no xact should restart when we are idle
	if (isIdle)
		return;

	// retry immediately if needed and possible
	// XXX: request type may be different this time
	if (retry && tryLaunch(genXact(oid)))
		return;
		
	// push waiting xactions forward
	if (theLaunchDebts.count() && !theConnMgr->atHardConnLimit())
		resumeXact();
}

CltXact *Client::fetch(const ObjId &oid, DistrPoint *dp) {
	Assert(oid);
	CltXact *x = genXact(oid);
	x->cacheDistrPoint(dp);
	if (tryLaunch(x))
		return x;
	return 0;
}

// must not be called directly but rather through tryLaunch
bool Client::launch(CltXact *x) {
	Assert(x);
	Assert(!theConnMgr->atHardConnLimit());

	const NetAddr &connAddr = x->nextHop();
	Assert(connAddr);
	Assert(isdigit(connAddr.addrA()[0]));

	if (Connection *conn = theConnMgr->get(connAddr)) {
		theCcXactLvl++;
		x->exec(this, conn);
		return true;
	}

	if (ReportError(errConnectEstb))
		Comment(1) << theHost << " failed to connect to " << connAddr << endc;

	launchFailed(x);
	return false;
}

// tries to generate and launch a transaction
// returns true if the xaction will be launched [eventually]
bool Client::tryLaunch() {
	if (theExtraLaunchLvl) {
		--theExtraLaunchLvl;
		return false;
	}
	
	return tryLaunch(genXact());
}

// will launch the xaction (if possible) or postpone it (if not)
// returns true if the xaction will be launched [eventually]
bool Client::tryLaunch(CltXact *x) {
	Assert(x);

	if (isIdle) {
		launchFailed(x);
		return false;
	}

	if (!x->nextHop()) {

		// should we ask peers for the best server?
		// XXX: remove askedPeer; the x->nextHop() should be enough?
		if (theCfg->thePeerIcp && !x->askedPeer()) {
			askPeer(theCfg->thePeerIcp, x);
			return true;
		}

		// next hop address should be known at this time 
		// (but my not be resolved yet)
		if (!setNextHopAddr(x))
			return false;

		// should we lookup next hop address?
		if (theDnsMgr->needsLookup(x->nextHop())) {
			lookupAddr(x);
			return true;
		}
	}

	// check if we should postpone the xaction
	if (theConnMgr->atHardConnLimit()) {
		if (theCfg->theWaitXactLmt < 0 || theLaunchDebts.count() < theCfg->theWaitXactLmt) {
			suspendXact(x);
			return true;
		}

		if (ReportError(errTooManyWaitXact))
			Comment(3) << "xactions active: " << theCcXactLvl 
				<< " waiting: " << theLaunchDebts.count() 
				<< " limit: " << theCfg->theWaitXactLmt << endc;
		launchFailed(x);
		return false;
	}

	return launch(x);
}

void Client::suspendXact(CltXact *x) {
	theLaunchDebts.append(x);
	Broadcast(TheWaitBegChannel, x);
}

void Client::resumeXact() {
	CltXact *x = theLaunchDebts.pop();
	Broadcast(TheWaitEndChannel, x);
	launch(x);
}

void Client::launchFailed(CltXact *x) {
	Assert(x);
	if (!isIdle)
		Broadcast(TheXactErrChannel, x);
	x->noteAbort();
	TheXacts->put(x);
}

CltXact *Client::genXact(const ObjId &oid) {
	CltXact *x = TheXacts->get();
	x->oid(oid);
	return x;
}

CltXact *Client::genXact() {
	ObjId oid;
	genOid(oid);
	return genXact(oid);
}

void Client::genOid(ObjId &oid) {
	selectViserv(oid);
	selectTarget(oid);
	selectObj(oid);
	selectContType(oid);
	selectReqType(oid);
	selectReqMethod(oid);

	if (theCfg->genUniqUrls)
		oid.world(UniqId::Create()); // changes every time
}

void Client::selectViserv(ObjId &oid) {
	oid.viserv(theCfg->selectViserv());
	Assert(oid.viserv() >= 0);
}

void Client::selectTarget(ObjId &oid) {
	const int visName = TheViservs.nameIdxAt(oid.viserv());
	const int target = oid.type() < 0 ?
		selectAnyTarget(visName) :
		selectTypedTarget(oid, visName);

	// sanity check
	const HostCfg *hcfg = TheHostMap->at(target);
	Assert(hcfg->theServerRep);

	oid.target(target);
}
	
void Client::selectObj(ObjId &oid) {
	static RndGen rng;

	OidGenStat &oidGenStat = TheStatPhaseMgr->oidGenStat();

	const bool needPub = rng.event(theCfg->thePublicRatio);
	const bool needRepeat = rng.event(theCfg->theRecurRatio*
		TheStatPhaseMgr->recurFactor().current());
	oidGenStat.recordNeed(needRepeat, needPub);

	PrivWorld &privWorld = thePrivWorlds[oid.viserv()];
	PubWorld &pubWorld = *ThePubWorlds[oid.viserv()];

	const bool privCanRep = privWorld.canRepeat();
	const bool privCanProd = privWorld.canProduce();

	const bool pubCanRep = pubWorld.canRepeat();
	const bool pubCanProd = pubWorld.canProduce();

	const bool canRep = privCanRep || pubCanRep;
	const bool canProd = privCanProd || pubCanProd;

	// the logic gives priority to repeatOid goal rather than to genPubOid
	if (canRep && (needRepeat || !canProd)) {
		if (pubCanRep && (needPub || !privCanRep)) {
			pubWorld.repeat(oid, theCfg->thePopModel);
			oidGenStat.recordGen(true, true);
			return;
		}

		Assert(privCanRep); 
		privWorld.repeat(oid, theCfg->thePopModel);
		oidGenStat.recordGen(true, false);
		return;
	}

	// new public
	if (pubCanProd && (needPub || !privCanProd)) {
		pubWorld.produce(oid, rng);
		oidGenStat.recordGen(false, true);
		return;
	}

	// new private object (last resort, never fails)
	privWorld.produce(oid, rng);
	oidGenStat.recordGen(false, false);
}

void Client::selectContType(ObjId &oid) {
	Assert(oid.type() < 0); // we do not overwrite existing setting
	const HostCfg *hcfg = TheHostMap->at(oid.target());
	Assert(hcfg);
	Assert(hcfg->theContent);
	const ContentCfg *ccfg = hcfg->theContent->getDir(oid);
	oid.type(ccfg->id());
}

void Client::selectReqType(ObjId &oid) {
	static RndGen rng;
	const int reqType =
		rng.event(TheStatPhaseMgr->specialMsgFactor().current()) ?
		(int)theCfg->theReqTypeSel->trial() : rqtBasic;
	oid.ims200(reqType == rqtIms200);
	oid.ims304(reqType == rqtIms304);
	oid.reload(reqType == rqtReload);
	oid.rediredReq(false);
}

void Client::selectReqMethod(ObjId &oid) {
	static RndGen rng;
	const int reqMethod =
		rng.event(TheStatPhaseMgr->specialMsgFactor().current()) ?
		(int)theCfg->theReqMethodSel->trial() : rqmGet;
	oid.get(reqMethod == rqmGet);
	oid.post(reqMethod == rqmPost);
	oid.head(reqMethod == rqmHead);
	oid.put(reqMethod == rqmPut);
}

// find any target behind a visible name
int Client::selectAnyTarget(int visName) {
	const NetAddr &addr = TheAddrMap->selectAddr(visName);
	int idx = -1;
	Assert(TheHostMap->find(addr, idx));
	return idx;
}

// find a target that has requested oid type
int Client::selectTypedTarget(ObjId &oid, int visName) {
	Assert(oid.type() > 0);
	for (AddrMapAddrIter i = TheAddrMap->addrIter(visName); i; ++i) {
		int idx = -1;
		Assert(TheHostMap->find(i.addr(), idx));
		const HostCfg *hcfg = TheHostMap->at(idx);
		Assert(hcfg->theContent);
		if (hcfg->theContent->hasContType(oid.type()))
			return idx;
	}

	// we failed to find a target that has the right content
	if (ReportError(errUnreachContType)) {
		const NetAddr &srvAddr = TheViservs.nameAt(visName);
		const String kind = TheContentMgr.get(oid.type())->kind() ?
			TheContentMgr.get(oid.type())->kind() : String("undefined");
		Comment << "robot at " << host() << " cannot find content of '"
			<< kind << "' kind on server(s) visible as " << srvAddr << endc;
	}

	return selectAnyTarget(visName); // should select new visName instead?
}
	
void Client::noteRedirect(CltXact *, const ObjId &oid) {
	Assert(oid.rediredReq());
	if (tryLaunch(genXact(oid)))
		theExtraLaunchLvl++;
}

void Client::noteEmbedded(CltXact *, const ObjId &oid) {
	static RndGen rng;
	if (!rng.event(theCfg->theEmbedRecurRatio))
		return;

	if (tryLaunch(genXact(oid)))
		theExtraLaunchLvl++;
}

void Client::askPeer(const NetAddr &addr, CltXact *x) {
	IcpCltXact *q = TheIcpXacts.get();
	q->reason(x);
	q->exec(this, addr);
}

void Client::notePeerAsked(IcpCltXact *q) {
	Assert(q);

	CltXact *x = q->reason(this);
	x->usePeer(q->hit());
	/* pass ICP stats here ... */

	TheIcpXacts.put(q);

	tryLaunch(x);
}

void Client::lookupAddr(CltXact *x) {
	theDnsMgr->lookup(x->nextHop(), x);
}

void Client::noteAddrLookup(const NetAddr &addr, CltXact *x) {
	Assert(x);

	// successful lookup?
	if (addr) {
		x->nextHopVar() = addr; // update
		tryLaunch(x);
	} else {
		launchFailed(x);
	}
}

// select the address for the next hop connection
bool Client::setNextHopAddr(CltXact *x) const {
	NetAddr &addr = x->nextHopVar();

	if (x->usePeer() && theCfg->thePeerHttp)
		return addr = theCfg->thePeerHttp;

	if (theProxyAddr)
		return addr = theProxyAddr;

	if (const char *uri = x->oid().foreignUrl()) {
		if (!MsgHdr::ParseHostInUri(uri, uri+x->oid().foreignUrl().len(), addr)) {
			ReportError(errHostlessForeignUrl);
			return false;
		}
		return addr;
	}

	return addr = TheViservs.nameAt(x->oid().viserv());
}

PortMgr *Client::getPortMgr(bool bind2iface) {
	const NetAddr addr = bind2iface ? theHost : NetAddr();

	// check if we already have a port mgr for host
	for (int i = 0; i < ThePortMgrs.count(); ++i) {
		if (ThePortMgrs[i]->addr() == addr)
			return ThePortMgrs[i];
	}

	// create new port manager
	PortMgr *mgr = TheCltOpts.thePorts.set() ?
		(PortMgr*) new ExpPortMgr(addr, TheCltOpts.thePorts.lo(), TheCltOpts.thePorts.hi()) :
		(PortMgr*) new EphPortMgr(addr);
	ThePortMgrs.append(mgr);
	return mgr;
}


int Client::logCat() const {
	return lgcCltSide;
}

void Client::FreezeSrvWss() {
	for (int i = 0; i < ThePubWorlds.count(); ++i)
		ThePubWorlds[i]->freezeWss();
}

void Client::Farm(XactFarm<CltXact> *aFarm) {
	Assert(!TheXacts && aFarm);
	TheXacts = aFarm;
}

void Client::LogState(OLog &log) {
	log << bege(lgSrvRepState, lgcCltSide);
	OLogStorePtrs(log, ThePubWorlds);
	log << ende;
}
