
/*
 * Web Polygraph                        http://www.web-polygraph.org/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1998-2000 by The Regents of the
 *  University of California. See the COPYRIGHT file for details.
 */

#include "pgl/pgl.h"

#include "xstd/h/strstream.h"

#include "xstd/String.h"
#include "xstd/gadgets.h"
#include "pgl/PglStringSym.h"
#include "pgl/PglArraySym.h"
#include "pgl/PglStrRangeLexer.h"
#include "pgl/PglStrBlocks.h"
#include "pgl/PglStrRange.h"


PglStrRange::PglStrRange() {
}

PglStrRange::PglStrRange(const PglStrRange &r) {
	theBlocks.stretch(r.theBlocks.count());
	for (int i = 0; i < r.theBlocks.count(); ++i)
		theBlocks.append(r.theBlocks[i]->clone());
	theImage = r.theImage;
}

PglStrRange::~PglStrRange() {
	reset();
}

void PglStrRange::reset() {
	while(theBlocks.count()) delete theBlocks.pop();
	theImage = 0;
}

int PglStrRange::count() const {
	int cnt = theBlocks.count() ? 1 : 0;
	for (int i = 0; i < theBlocks.count(); ++i)
		cnt *= theBlocks[i]->count();
	return cnt;
}

int PglStrRange::iterPos() const {
	int pos = 0;
	int cnt = 1;
	for (int i = theBlocks.count()-1; i >= 0; --i) {
		pos += theBlocks[i]->pos()*cnt;
		cnt *= theBlocks[i]->count();
	}
	return pos;
}

void PglStrRange::iterate(Iter iter) const {
	startIter();
	do {
		String str;
		currentIter(str);
		iter(str);
	} while (nextIter());
}

String PglStrRange::toStr() const {
	ostrstream buf;
	print(buf);
	buf << ends;

	const String res = buf.str();
	ostrstreamFreeze(buf, false);
	return res;
}

void PglStrRange::toStrs(Array<String*> &strs) const {
	startIter();
	do {
		String str;
		currentIter(str);
		strs.append(new String(str));
	} while (nextIter());
}

ArraySym *PglStrRange::toSyms(const TokenLoc &loc) const {
	ArraySym *arr = new ArraySym("string");

	startIter();
	do {
		String str;
		currentIter(str);
		StringSym ss(str);
		if (loc)
			ss.loc(loc);
		arr->add(ss); // XXX: use range!
	} while (nextIter());

	return arr;
}

void PglStrRange::strAt(int idx, String &str) const {
	startIter();
	skipIter(idx);
	currentIter(str);
}

void PglStrRange::startIter() const {
	for (int i = 0; i < theBlocks.count(); ++i)
		theBlocks[i]->start();
}

void PglStrRange::skipIter(int count) const {
	count -= iterPos();
	Assert(count >= 0);
	int right = 1;
	int check = 0;
	for (int i = theBlocks.count()-1; count > check && i >= 0; --i) {
		PglStrBlock &b = *theBlocks[i];
		const int pos = (count/right) % b.count();
//clog << here << i << ": " << "pos: " << pos << " == " << '(' << count << '/' << right << ") % " << b.count() << endl;
		b.pos(pos);
		check += pos*right;
		right *= b.count();
	}
	Assert(count == check);
}

bool PglStrRange::nextIter() const {
	return nextIter(theBlocks.count()-1);
}

bool PglStrRange::nextIter(int level) const {
	if (level < 0)
		return false;

	PglStrBlock &b = *theBlocks[level];
	// delegate to next level if overflow
	if (b.atLast()) {
		b.start();
		return nextIter(level-1);
	}

	b.next();
	return true;
}

void PglStrRange::currentIter(String &str) const {
	ostrstream buf;
	for (int i = 0; i < theBlocks.count(); ++i) {
		theBlocks[i]->printCur(buf);
	}
	buf << ends;
	str = buf.str();
	ostrstreamFreeze(buf, false);
}

bool PglStrRange::parse(const String &val) {
	theImage = val;
	PglStrRangeLexer lexs(this, theImage);

	// convert val into lexems
	if (!lexs.parse())
		return false;

	// convert lexems into blocks
	while (lexs) {
		if (escapeChar(lexs.next()) && lexs.next(1)) {
			lexs.skip();
			lexs.step();
		} else
		if (optRangeBeg(lexs.next())) {
			lexs.step(); // keep a separator
			(void)lexs.range(true); // optional
		} else
		if (reqRangeBeg(lexs.next())) {
			const char start = lexs.next();
			lexs.skip(); // skip the range separator
			if (lexs.range(false) && reqRangeEnd(start, lexs.next()))
				lexs.skip();
			else
				return false; // invalid range specs
		} else
		if (reqRangeEnd(lexs.next())) {
			return false; // unbalanced required range separator
		} else {
			lexs.step(); // simple label
		}
	}

	return true;
}

bool PglStrRange::escapeChar(char ch) const {
	return ch == '\\';
}

bool PglStrRange::optRangeBeg(char) const {
	return false;
}

bool PglStrRange::reqRangeBeg(char ch) const {
	return ch == '[';
}

bool PglStrRange::reqRangeEnd(char ch) const {
	return ch == ']';
}

bool PglStrRange::reqRangeEnd(char, char ch) const {
	return reqRangeEnd(ch);
}

void PglStrRange::addPointBlock(const char *start, const char *stop) {
	theBlocks.append(new PglStrPointBlock(start, stop));
}

void PglStrRange::addRangeBlock(int start, int stop, bool isolated) {
	theBlocks.append(new PglStrRangeBlock(start, stop, isolated));
}

bool PglStrRange::canMergeStrs(const PglStrRange &r) const {
	if (!theImage) // brand new object
		return true;

	if (theBlocks.count() != r.theBlocks.count())
		return false;

	int diffCount = 0;
	for (int i = 0; diffCount <= 1 && i < theBlocks.count(); ++i) {
		const PglStrBlock &b1 = *theBlocks[i];
		const PglStrBlock &b2 = *r.theBlocks[i];
		diffCount += b1.diffCount(b2);
	}
	// can merge ranges that differ by one block only because 
	// the range notation cannot express other resulting sets
	return diffCount == 1;
}

void PglStrRange::merge(const PglStrRange &r) {
	if (theImage) {
		Assert(theBlocks.count() == r.theBlocks.count());
		for (int i = 0; i < theBlocks.count(); ++i)
			theBlocks[i]->merge(*r.theBlocks[i]);
		theImage = toStr();
	} else {
		// brand new object
		Assert(parse(r.toStr()));
	}

}

int PglStrRange::rangedBlockCnt() const {
	int count = 0;
	for (int i = 0; i < theBlocks.count(); ++i) {
		if (theBlocks[i]->count() > 1)
			count++;
	}
	return count;
}

ostream &PglStrRange::print(ostream &os) const {
	for (int i = 0; i < theBlocks.count(); ++i)
		theBlocks[i]->print(os);
	return os;
}

